/*

	psfix:

	This program massages the output of psdit into the desired
	format for the POSIX drafts and the BCS's.  It corrects some
	troff deficiencies/bugs in handling line numbers and diff
	marks.  (Note:  the problem with diff marks occurs only when
	they are used on the same line as a line number; they weave
	around the right margin without lining up.)  If also integrates
	PostScript documents or figures into the TranScript file
	produced by troff.

	The troff input had to have been prepared in a certain way:

		1.  The line numbers have to be in font 6.  An easy way
		to do this is to use a macros ".Nm" for your numbering,
		defined as such:

			.de Nm
			.br
			.S 8
			.ft 6
			.nm \\$1 \\$2 \\$3
			.ft 1
			.S
			..

		That also makes them a comfortable point size 8.  The
		main processing with line numbers is to avoid
		overwritings that normally occurs with tbl or eqn.
		When it writes a line number it ensures that it hasn't
		written one less than 90 pixels above it.

		2.  The diff marks have to be point size 8.  This
		program takes care of diffmarks 7-9 and A-C, but it is
		very easy to modify it; just search for "789ABC".  To
		get the diff marks to be 8, you can use a sed script:

			sed -e 's/^\(.\)mc 9/\1mc \\f1\\s89\\s0\\fP/' \  
			    -e 's/^\(.\)mc a/\1mc \\f1\\s8A\\s0\\fP/' 

		piped after the tbl process and before troff.  The
		program lines all the diff marks up at pixel 4375,
		which can also be changed.

		3.  Blank lines in tables that want diff marks (to show
		deleted lines) need to be entered as "JunK" on a line
		by itself.  tbl can't seem to print the diffmark on the
		correct line unless you do this.

		4.  Figures are called in with the .fG macro, as follows:

			.fG <space to reserve> <pathname of figure>

		5.  Shaded margins can be accomplished with the .rS/.rE
		(start/stop Rationale) macro pair.

		

						Hal Jespersen
						November, 1988

 */
#include <stdio.h>
/*#include <string.h>, does not exist on the Pyramid */
#include <string.h>
#include <memory.h>
/*#include <memory.h>, does not exist on the Pyramid */

extern int	optind;
extern char	*optarg;
#define MARKSPOT	4375	/* pixel spot for diff marks */
char	buf[1024];	/* main buffer for I/O */

main(argc, argv)
int	argc;
char	*argv[];
{

	int	s8, f6, infig, x, xpos, ypos, shading, shade_y;
	int	currenty, prevynum, prevymark, lineno, lastmark;
	char	marks[16];
	char	tag[2];
	char	figname[256];
	char	*str, *findstring();

	f6 = s8 = infig = shading = shade_y = 0;
	prevynum = prevymark = 0;
	lastmark = 0;

	while (gets(buf) != NULL) {
		if (strncmp(buf, "%%Page: ", 8) == 0) {
			/* If a new page, reset all our flags */
			f6 = s8 = 0;
			prevynum = prevymark = 0;
			puts(buf);
			continue;
		}
		if (strncmp(buf, "8 s", 3) == 0) {
			/* Entering point size 8 */
			s8 = 1;
			puts(buf);
			continue;
		}
		if (sscanf(buf, "%d %[s]", &xpos, tag) == 2) {
			/* Entering some other point size */
			s8 = 0;
			puts(buf);
			continue;
		}
		if (strncmp(buf, "6 f", 3) == 0) {
			/* Entering font position 6 */
			f6 = 1;
			printf("1 f\n");
			continue;
		}
		if (strncmp(buf+1, " f", 2) == 0) {
			/* Entering some other font position */
			f6 = 0;
			puts(buf);
			continue;
		}
		if (sscanf(buf, "%d %d(%d)%[N]", &xpos, &ypos, &x, tag) == 4) {
			/* Keep track of latest vertical position of line numbers;
			   this is done in case embedded figures are getting
			   distracted with line numbers. */
			currenty = ypos;
		}
		if (sscanf(buf, "%d %[p]", &x, tag) == 2) {
			/* end of page; check shading */
			if (shading) {
				shade(shade_y, 5650);
					/* shade from previous start to end
					   of page */
				shade_y = 600;
					/* set shade-start to top of next page */
			}
			puts(buf);
			continue;
		}
		if (sscanf(buf, "%d %d(HLJ_RAT_S)%[N]", &xpos, &ypos, tag) == 3) {
			/* Found start of Rationale to be shaded. */
			if (!shading) {
				/* test if we're already shading, which is OK */
				shade_y = ypos;	/* save starting position */
				shading = 1;
			}
			continue;
		}
		if (sscanf(buf, "%d %d(HLJ_RAT_E)%[N]", &xpos, &ypos, tag) == 3) {
			/* Found end of shaded area. */
			if (shading) {
				/* guard against lone end blocks */
				shade(shade_y, ypos);
			}
			shading = 0;
			shade_y = 0;
			continue;
		}
		if (sscanf(buf, "%d %d(HLJ_FIG)%[N]", &xpos, &ypos, tag) == 3) {
			/* Found start of embedded PostScript figure */
			/* This is the type of figure header found when it
			   isn't done in a diversion */
			infig = 1;
			currenty = ypos;
			continue;
		}
		if (sscanf(buf, "%d(HLJ_FIG)%[X]", &xpos, tag) == 2) {
			/* Found (alternative) start of embedded PostScript figure */
			infig = 1;
			continue;
		}
		if (infig && !s8) {
			/* do the figure insertion.  The pathname is embedded
			   into the text; the two styles are in or out of
			   diversions (with/without line numbers) */
			if (sscanf(buf, "%d %d(%[a-zA-Z_0-9/])%[N]", &xpos, &ypos, figname, tag) == 4) {
				readfig(figname, 0, ypos);
				infig = 0;
				continue;
			}
			if (sscanf(buf, "%d(%[a-zA-Z_0-9/])%[X]", &xpos, figname, tag) == 3) {
				readfig(figname, 0, currenty);
				infig = 0;
			}
			continue;
		}
		if (f6) {
			/* Process possible line numbers within font 6 */
			if (sscanf(buf, "%d %d(%d)%[N]", &xpos, &ypos, &lineno, tag) == 4) {
				if (ypos < prevynum) {
					printf("%d %d( )N\n", 0, ypos);
				} else {
					puts(buf);
					prevynum = ypos + 90;
						/* ensure next number is at least
						   90 units farther down */
				}
				continue;
			}
			if (sscanf(buf, "%d(%d)%[X]", &xpos, &lineno, tag) == 3) {
				continue;
			}
			/* Everything else passes through ... */
		}
		if (s8) {
			/* Process possible diff marks within size 8 */
			marks[0] = marks[1] = marks[2] = 0;
			if (sscanf(buf, "%d(%[789ABC])%[X]", &xpos, marks, tag) == 3
			    && testpos(xpos))
				goto putmark;
			if (sscanf(buf, "%d %d(%[789ABC])%[N]", &xpos, &ypos, marks, tag) == 4
			    && testpos(xpos)
			    && marks[2] == 0) {
				/* for the single mark with N, blank it out,
				   but still print it to get the proper vertical
				   spacing */
				printf("%d %d( )N\n", xpos, ypos);
				continue;
			}
			if (sscanf(buf, "%d %d(%[789ABC])%[X]", &xpos, &ypos, marks, tag) == 4
			    && testpos(xpos)
			    && marks[2] == 0)
				goto putmark;
			if (sscanf(buf, "%d %d(%[789ABC])%[N]", &xpos, &ypos, marks, tag) == 4
				/* Junk these */
			    && testpos(xpos)
			    && marks[2])
				continue;
			if (sscanf(buf, "(%[789ABC])%[S]", marks, tag) == 2
			    && strchr("789ABC", marks[0]))
				goto putmark;

			/* No match; pass it through and reset flag that
			   would have shown that we put out a diff mark. */
			puts(buf);
			lastmark = 0;
			continue;
putmark:
			/* Common spot to write all the diff marks in the
			   form "4375(m)X". */
			if (!lastmark) {
				printf("%d(%c)X\n", MARKSPOT, marks[0]);
				lastmark = 1;
			}
			continue;
		}

		/* Blank out the JunK entries. */
		if ((str = findstring(buf, "JunK")) != NULL)
			memcpy(str, "    ", 4);
		puts(buf);
		lastmark = 0;

		/* If we've run into a table with rules, force on lastmark.
		   This will prevent us from thinking we have a clear
		   spot to write in.  Normally, the rules are so thin that
		   two diffmarks overlay each other. */
		if (findstring(buf, "hhhhh")
		   || findstring(buf, "iiiii")
		   || findstring(buf, "(c)Y"))
			lastmark = 1;
	}
	if (ferror(stdin)) {
		fprintf(stderr, "Read error on input file\n");
		exit(1);
	}
	exit(0);
}

/* Return success if the horizontal position is in about the right
   range for a diff mark. */
testpos(pos)
int	pos;
{
	if (pos < 3500 || pos > 4500)
		return 0;
	return 1;
}

/* See if string s2 is wholly contained in s1.  Return a pointer to the
   first instance, if so.  If not, return NULL. */
char *
findstring(s1, s2)
char	*s1, *s2;
{
	int	n = strlen(s2);

	while (*s1) {
		if (*s1 != *s2) {
			s1++;
			continue;
		}
		if (strncmp(s1, s2, n) == 0)
			return s1;
		s1++;
	}
	return NULL;
}

readfig(file, argx, argy)
char	*file;
int	argx, argy;
{
	int	llx, lly, urx, ury;
	int	newllx, newlly, newurx, newury;
	FILE	*fig;

	if ((fig = fopen(file, "r")) == NULL ){
		fprintf(stderr, "Can't open %s\n", file);
		exit(1);
	}

	argy -= 150;	/* kludge to avoid extra space at top of figure */

	printf("/HLJ save def\n");
	printf("/locate {%% LLx LLy URx URy newLLx newLLy newURx newURy\n");
	printf(" 8 dict begin\n");
	printf("  [/newURy /newURx /newLLy /newLLx /URy /URx /LLy /LLx]\n");
	printf("  { exch def } forall\n");
	printf("  newLLx newLLy translate %% move origin to new location\n");
	printf("  newURx newLLx sub URx LLx sub div %% X scale\n");
	printf("  newURy newLLy sub URy LLy sub div scale %% Y scale\n");
	printf("  LLx neg LLy neg translate %% compensate for LL offset\n");
	printf(" end\n");
	printf("} bind def\n");

	while (fgets(buf, 1024, fig)) {
		buf[strlen(buf) - 1] = 0;	/* remove fgets' trailing \n */
		if (sscanf(buf, "%%%%BoundingBox:%d %d %d %d", &llx, &lly, &urx, &ury) == 4) {
			newllx = argx;
			newlly = argy;
			if (newllx == 0)
				newllx = (72*8.5/2 - (urx - llx)/2) * 8;
			newurx = (newllx/8 + (urx - llx)) * 8;
			newury = (newlly/8 - (ury - lly)) * 8;
			printf("%d %d %d %d %d %d %d %d locate\n",
				llx, lly, urx, ury,
				newllx, newlly, newurx, newury);
		}
		if (strncmp(buf, "%%Trailer", 9) != 0)
			puts(buf);
	}
	if (ferror(fig)) {
		fprintf(stderr, "read error on input file %s\n", file);
	}
	printf("HLJ restore\n");
	fclose(fig);
}

/* shade() prints shading margin bars for Rationale and other parts of
   text not in the standard.  The args are the Y position for the start
   and end of the marks.
 */
shade(ys, ye)
int ys, ye;
{
	printf("/HLJ save def\n");
	printf("%d MX\n", 400);
	printf("%d MY\n", ys);
	printf("200 setlinewidth\n");
	printf(".90 setgray\n");
	printf("0 %d Dl\n", ye - ys);	/* print left bar */
	printf("%d MX\n", 4550);
	printf("%d MY\n", ys);
	printf("0 %d Dl\n", ye - ys);	/* print right bar */
	printf("HLJ restore\n");
}
